/**
 * @file ClientHandler.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class ClientHandler.
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using System.IO;

using Objects3D;
using GRV11;

namespace Cubes3D
{
    /** 
     * @class ClientHandler
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for all functionality concerning the client.
     */
    public class ClientHandler
    {
        private static Color FBackgroundColor  = Color.Black;
        private static Color FDefaultCubeColor = Color.LightGray;
        private static Color FFirstCubeColor   = Color.LightGreen;
        private static Color FLastCubeColor    = Color.LightSalmon;
        private static Color FPlayerColor      = Color.LightGoldenrodYellow;

        private Form FOwner = null;

        private Controller FController = null;
        private Cube[]     FCubes      = null;
        private Player[]   FPlayers    = null;

        private Coordinate FReplayPos;

        private GRControl FGRControl = null;
        private Thread FClientReceiver = null;
        private UdpClient FClientSender = null;
        private IPEndPoint FServerEndPoint = null;

        private bool FActive = false;
        private bool FKill   = false;

        private bool   FPlayReplay     = false;
        private bool   FRecordReplay   = true;
        private String FReplayFileName = "";

        private int FClientPort = 9999;
        private int FServerPort = 8888;

        private int FTimeStamp = 0;

        private bool FAcknowledged = false;

        private bool FLocal = false;

        private bool FEnded = false;
        private bool FFinished = false;

        private int FGameStart = 0;
        private int FGameStop = 0;

        public ClientHandler(Form AOwner, GRControl AGRControl, bool ALocal)
        {
            FLocal = ALocal;

            FOwner     = AOwner;
            FGRControl = AGRControl;

            FController = new Controller();
            FController.ID = 0;

            FReplayPos = new Coordinate(0, 0, 0);

            FReplayFileName = System.Environment.CurrentDirectory + "\\Replay.rpl";

            FClientReceiver = new Thread(new ThreadStart(StartClientReceiver));
            FClientReceiver.IsBackground = false; // simplify thread killing
        }

        /**
         * @method SetLight
         * 
         * @author Kevin Niehage
         * 
         * @param[in] GR instance
         * @param[in] light position
         * 
         * @brief Activates the lighting.
         */
        private void SetLight(GR AGR, Coordinate APosition)
        {
            if ((AGR != null) && (APosition != null))
            {
                ///*
                AGR.glLightfv(GR.GL_LIGHT0, GR.GL_AMBIENT, new float[] { 0.3F, 0.3F, 0.3F, 0.0F });
                AGR.glLightfv(GR.GL_LIGHT0, GR.GL_DIFFUSE, new float[] { 1.0F, 1.0F, 1.0F, 0.0F });
                //*/

                AGR.glLightfv(GR.GL_LIGHT0, GR.GL_POSITION, new float[] { - APosition.X, - APosition.Y + 150.0F, - APosition.Z, 0.0F });
                AGR.glLightfv(GR.GL_LIGHT0, GR.GL_SPOT_DIRECTION, new float[] { 0, 0, -1 });
                AGR.glLightf(GR.GL_LIGHT0, GR.GL_CONSTANT_ATTENUATION, 0.00003F);
                AGR.glLightf(GR.GL_LIGHT0, GR.GL_LINEAR_ATTENUATION, 0.00003F);
                AGR.glLightf(GR.GL_LIGHT0, GR.GL_QUADRATIC_ATTENUATION, 0.00003F);
                AGR.glEnable(GR.GL_LIGHT0);

                AGR.glEnable(GR.GL_LIGHTING);
            }
        }

        /**
         * @method StartClientReceiver
         * 
         * @author Kevin Niehage
         * 
         * @brief Main method of the client UDP listener thread.
         */
        private void StartClientReceiver()
        {
            IPEndPoint LClientEndPoint = null;
            if (FLocal)
            {
                LClientEndPoint = new IPEndPoint(IPAddress.Loopback, FClientPort);
            }
            else
            {
                LClientEndPoint = new IPEndPoint(IPAddress.Any, FClientPort);
            }

            UdpClient LClientReceiver = new UdpClient(FClientPort);
            LClientReceiver.EnableBroadcast = (!(FLocal));

            byte[] LReceivedData;

            if (FLocal)
            {
                FServerEndPoint = new IPEndPoint(IPAddress.Loopback, FServerPort);
            }
            else
            {
                FServerEndPoint = new IPEndPoint(IPAddress.Broadcast, FServerPort);
            }

            FClientSender = new UdpClient();
            FClientSender.EnableBroadcast = (!(FLocal));

            ASCIIEncoding LEncoding = new ASCIIEncoding();

            bool LSave = false;

            float LRadX = 0.0F;
            float LRadY = 0.0F;
            FileStream LReplayFile = null;

            byte[] LByteHeader = new byte[3];
            byte[] LByteStamp  = new byte[sizeof(int)];
            int    LFileStamp  = -1;
            int    LLastTick   = 0;
            byte   LLength     = 0;
            int    LTimeStamp  = -1;

            int  LAck    = 0;
            bool LNew    = false;
            int  LRandom = new Random().Next();

            try
            {
                while (!(FKill))
                {
                    if (FActive)
                    {
                        LLastTick = System.Environment.TickCount;

                        if (FPlayReplay)
                        {
                            FAcknowledged = true;

                            if (LReplayFile == null)
                            {
                                if (File.Exists(FReplayFileName))
                                {
                                    LReplayFile = File.OpenRead(FReplayFileName);
                                }
                            }

                            if (LReplayFile != null)
                            {
                                LRadX = Coordinate.DegToRad(FController.Rotation.X);
                                LRadY = Coordinate.DegToRad(FController.Rotation.Y);

                                if (FController.GetKeyState(ControllerKey.KEY_DOWN))
                                {
                                    FReplayPos.X += 5.0F * (float)Math.Sin(LRadY);
                                    FReplayPos.Y -= 5.0F * (float)Math.Sin(LRadX);
                                    FReplayPos.Z -= 5.0F * (float)Math.Cos(LRadY);
                                }
                                if (FController.GetKeyState(ControllerKey.KEY_LEFT))
                                {
                                    FReplayPos.X += 2.5F * (float)Math.Cos(LRadY);
                                    //FReplayPos.Y += 0.0F * (float)Math.Sin(LRadX);
                                    FReplayPos.Z += 2.5F * (float)Math.Sin(LRadY);
                                }
                                if (FController.GetKeyState(ControllerKey.KEY_RIGHT))
                                {
                                    FReplayPos.X -= 2.5F * (float)Math.Cos(LRadY);
                                    //FReplayPos.Y += 0.0F * (float)Math.Sin(LRadX);
                                    FReplayPos.Z -= 2.5F * (float)Math.Sin(LRadY);
                                }
                                if (FController.GetKeyState(ControllerKey.KEY_UP))
                                {
                                    FReplayPos.X -= 5.0F * (float)Math.Sin(LRadY);
                                    FReplayPos.Y += 5.0F * (float)Math.Sin(LRadX);
                                    FReplayPos.Z += 5.0F * (float)Math.Cos(LRadY);
                                }

                                LReplayFile.Read(LByteHeader, 0, LByteHeader.Length);
                                if (LEncoding.GetString(LByteHeader).Equals(Network.PlrHeader) || LEncoding.GetString(LByteHeader).Equals(Network.WldHeader))
                                {
                                    LReplayFile.Read(LByteStamp, 0, sizeof(int));
                                    LFileStamp = BitConverter.ToInt32(LByteStamp, 0);
                                    if (LTimeStamp < 0)
                                    {
                                        LTimeStamp = LFileStamp;
                                    }

                                    LLength = (byte)LReplayFile.ReadByte();

                                    LReplayFile.Seek(- sizeof(int) - 4, SeekOrigin.Current);
                                    if (LEncoding.GetString(LByteHeader).Equals(Network.PlrHeader))
                                    {
                                        LReceivedData = new byte[(Player.ByteSize * LLength) + (sizeof(int) * 1) + (sizeof(byte) * 4)];
                                    }
                                    else
                                    {
                                        LReceivedData = new byte[(Cube.ByteSize * LLength) + (sizeof(int) * 1) + (sizeof(byte) * 4)];
                                    }

                                    LReplayFile.Read(LReceivedData, 0, LReceivedData.Length);

                                    Network.SetCubes(LReceivedData, ref FCubes, ref FTimeStamp);
                                    Network.SetPlayers(LReceivedData, ref FPlayers, ref FTimeStamp);

                                    /*
                                    while (LTimeStamp < FTimeStamp)
                                    {
                                        LTimeStamp += (System.Environment.TickCount - LLastTick);
                                        LLastTick = System.Environment.TickCount;
                                    }
                                    */
                                }
                            }
                        }
                        else
                        {
                            LReceivedData = LClientReceiver.Receive(ref LClientEndPoint);
                            LReceivedData = Convert.FromBase64String(LEncoding.GetString(LReceivedData));

                            if (FAcknowledged)
                            {
                                if (Network.SetEnd(LReceivedData, ref LTimeStamp))
                                {
                                    FEnded = true;
                                }

                                byte LID = 0;
                                if (Network.SetFinish(LReceivedData, ref LID, ref LTimeStamp))
                                {
                                    FFinished = (LID == FController.ID);
                                    if (FFinished)
                                    {
                                        if (FGameStop == 0)
                                        {
                                            FGameStop = System.Environment.TickCount;
                                        }
                                    }
                                }

                                if ((FEnded || FFinished))
                                {
                                    FActive = false;
                                }

                                LSave = Network.SetCubes(LReceivedData, ref FCubes, ref FTimeStamp);
                                LSave = LSave || Network.SetPlayers(LReceivedData, ref FPlayers, ref FTimeStamp);

                                if ((LSave) && (FGameStart == 0))
                                {
                                    FGameStart = System.Environment.TickCount;
                                }

                                if ((LSave) && (FRecordReplay))
                                {
                                    File.AppendAllText(FReplayFileName, LEncoding.GetString(LReceivedData));
                                }
                            }
                            else
                            {
                                if (LNew)
                                {
                                    byte LID = 0;
                                    if (Network.SetAcknowledge(LReceivedData, ref LAck, ref LID, ref LTimeStamp))
                                    {
                                        FAcknowledged = (LAck == LRandom);
                                        if (FAcknowledged)
                                        {
                                            FController.ID = LID;
                                        }
                                    }

                                    if (!(FAcknowledged))
                                    {
                                        byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetEnter(LRandom, FTimeStamp)));
                                        FClientSender.Send(LData, LData.Length, FServerEndPoint);
                                    }
                                }
                                else
                                {
                                    LNew = Network.SetNew(LReceivedData, ref LTimeStamp);
                                    if (LNew)
                                    {
                                        byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetEnter(LRandom, FTimeStamp)));
                                        FClientSender.Send(LData, LData.Length, FServerEndPoint);
                                    }
                                }
                            }
                        }

                        if ((FAcknowledged) && (LSave))
                        {
                            if ((FOwner != null) && (FGRControl != null))
                            {
                                FOwner.Invoke(new EventHandler(delegate(object sender2, EventArgs args2)
                                {
                                    FGRControl.Invalidate();
                                }));
                            }
                        }

                        LTimeStamp += (System.Environment.TickCount - LLastTick);
                    }
                }
            }
            finally
            {
                LClientReceiver = null;
            }
        }

        /**
         * @property Active
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the client is working.
         */
        public bool Active
        {
            get
            {
                return FActive;
            }

            set
            {
                FActive = value;

                if (!(FClientReceiver.IsAlive))
                {
                    FClientReceiver.Start();
                }
            }
        }

        /**
         * @property Acknowledged
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether we got acknowledged by the server.
         */
        public bool Acknowledged
        {
            get
            {
                return FAcknowledged;
            }
        }

        /**
         * @property Ended
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the game has ended.
         */
        public bool Ended
        {
            get
            {
                return FEnded;
            }
        }

        /**
         * @property Finished
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether we reached the finish.
         */
        public bool Finished
        {
            get
            {
                return FFinished;
            }
        }

        /**
         * @property GameStart
         * 
         * @author Kevin Niehage
         * 
         * @brief Property holding the first tick count of the game.
         */
        public int GameStart
        {
            get
            {
                return FGameStart;
            }
        }

        /**
         * @property GameStop
         * 
         * @author Kevin Niehage
         * 
         * @brief Property holding the last tick count of the game.
         */
        public int GameStop
        {
            get
            {
                return FGameStop;
            }
        }

        /**
         * @property Kill
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the thread is about to be killed
         */
        public bool Kill
        {
            get
            {
                return FKill;
            }

            set
            {
                FKill = value;
            }
        }

        /**
         * @property PlayReplay
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether a replay is shown.
         */
        public bool PlayReplay
        {
            get
            {
                return FPlayReplay;
            }

            set
            {
                FPlayReplay = true;
                //FPlayReplay = value;
            }
        }

        /**
         * @property RecordReplay
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether a replay is recorded.
         */
        public bool RecordReplay
        {
            get
            {
                return FRecordReplay;
            }

            set
            {
                FRecordReplay = value;
            }
        }

        /**
         * @property ReplayFileName
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining where a replay is recorded to.
         */
        public String ReplayFileName
        {
            get
            {
                return FReplayFileName;
            }

            set
            {
                FReplayFileName = value;
            }
        }

        public void OpenGLStarted(GRControl AGRControl)
        {
            // empty
        }

        /**
         * @method Paint
         * 
         * @author Kevin Niehage
         * 
         * @param[in] event emitter
         * @param[in] event arguments
         * 
         * @brief Contains the whole drawing process
         */
        public void Paint(object sender, PaintEventArgs e)
        {
            if (FActive)
            {
                GRControl LGRControl = (sender as GRControl);
                if (LGRControl != null)
                {
                    GR LGR = LGRControl.GetGR();
                    if (LGR != null)
                    {
                        int LClientWidth = LGRControl.ClientRectangle.Width;
                        int LClientHeight = LGRControl.ClientRectangle.Height;
                        if (LClientWidth <= 0)
                        {
                            LClientWidth = 1;
                        }
                        if (LClientHeight <= 0)
                        {
                            LClientHeight = 1;
                        }

                        // Viewport - defines drawing range
                        LGR.glViewport(0, 0, LClientWidth, LClientHeight);

                        // Basic rendering conditions
                        LGR.glEnable(GR.GL_DEPTH_TEST);
                        LGR.glEnable(GR.GL_CULL_FACE);
                        LGR.glCullFace(GR.GL_BACK);
                        LGR.glDepthFunc(GR.GL_LEQUAL);
                        LGR.glFrontFace(GR.GL_CCW);

                        LGR.glClearColor(FBackgroundColor.R / 255.0F, FBackgroundColor.G / 255.0F, FBackgroundColor.B / 255.0F, FBackgroundColor.A / 255.0F);
                        LGR.glClear(GR.GL_COLOR_BUFFER_BIT | GR.GL_DEPTH_BUFFER_BIT);

                        LGR.glMatrixMode(GR.GL_PROJECTION);
                        LGR.glLoadIdentity();

                        LGR.gluPerspective(60.0f, ((double)(LClientWidth) / (double)(LClientHeight)), 0.1f, 64000f);

                        LGR.glMatrixMode(GR.GL_MODELVIEW);
                        LGR.glLoadIdentity();

                        // go to defaults
                        LGR.glRotatef(0f, 1f, 1f, 1f);
                        LGR.glScalef(1f, 1f, 1f);
                        LGR.glTranslatef(0f, 0f, 0f);

                        Player LLocalPlayer  = null;
                        Coordinate LPosition = new Coordinate(0.0F, 0.0F, 0.0F);

                        if (FPlayReplay)
                        {
                            LPosition = FReplayPos;
                        }
                        else
                        {
                            // find current player to set world accordingly
                            if (FPlayers != null)
                            {
                                foreach (Player LPlayer in FPlayers)
                                {
                                    if (LPlayer.ID == FController.ID)
                                    {
                                        LLocalPlayer = LPlayer;

                                        //SetLight(LGR, LPosition);

                                        FController.Rotation = LPlayer.Rotation;
                                        LPosition = LLocalPlayer.Position;
                                    }
                                }
                            }
                        }

                        LGR.glRotatef(FController.Rotation.X, 1.0F, 0.0F, 0.0F);
                        LGR.glRotatef(FController.Rotation.Y, 0.0F, 1.0F, 0.0F);
                        LGR.glRotatef(FController.Rotation.Z, 0.0F, 0.0F, 1.0F);

                        LGR.glTranslatef(LPosition.X, LPosition.Y, LPosition.Z);

                        if (FCubes != null)
                        {
                            LGR.glPushMatrix();
                            try
                            {
                                LGR.glColor4f(FDefaultCubeColor.R / 255.0F, FDefaultCubeColor.G / 255.0F, FDefaultCubeColor.B / 255.0F, FDefaultCubeColor.A / 255.0F);

                                foreach (Cube LCube in FCubes)
                                {
                                    if (LCube != null)
                                    {
                                        if (FCubes.Length >= 2)
                                        {
                                            if (LCube == FCubes[0])
                                            {
                                                LGR.glColor4f(FFirstCubeColor.R / 255.0F, FFirstCubeColor.G / 255.0F, FFirstCubeColor.B / 255.0F, FFirstCubeColor.A / 255.0F);

                                                LCube.Draw(LGR);

                                                LGR.glColor4f(FDefaultCubeColor.R / 255.0F, FDefaultCubeColor.G / 255.0F, FDefaultCubeColor.B / 255.0F, FDefaultCubeColor.A / 255.0F);
                                            }
                                            else
                                            {
                                                if (LCube == FCubes[FCubes.Length - 1])
                                                {
                                                    LGR.glColor4f(FLastCubeColor.R / 255.0F, FLastCubeColor.G / 255.0F, FLastCubeColor.B / 255.0F, FLastCubeColor.A / 255.0F);

                                                    LCube.Draw(LGR);

                                                    LGR.glColor4f(FDefaultCubeColor.R / 255.0F, FDefaultCubeColor.G / 255.0F, FDefaultCubeColor.B / 255.0F, FDefaultCubeColor.A / 255.0F);
                                                }
                                                else
                                                {
                                                    LCube.Draw(LGR);
                                                }
                                            }
                                        }
                                        else
                                        {
                                            LCube.Draw(LGR);
                                        }
                                    }
                                }
                            }
                            finally
                            {
                                LGR.glPopMatrix();
                            }
                        }

                        // draw players
                        IntPtr LQuadric = LGR.gluNewQuadric();
                        if (LQuadric != null)
                        {
                            try
                            {
                                if (FPlayers != null)
                                {
                                    foreach (Player LPlayer in FPlayers)
                                    {
                                        if (LPlayer != null)
                                        {
                                            if (LPlayer != LLocalPlayer)
                                            {
                                                LGR.glPushMatrix();
                                                try
                                                {
                                                    LGR.glColor4f(FPlayerColor.R / 255.0F, FPlayerColor.G / 255.0F, FPlayerColor.B / 255.0F, FPlayerColor.A / 255.0F);

                                                    LPlayer.Draw(LGR, LQuadric);
                                                }
                                                finally
                                                {
                                                    LGR.glPopMatrix();
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            finally
                            {
                                LGR.gluDeleteQuadric(LQuadric);
                            }
                        }

                        LGR.wglSwapBuffers(LGRControl.GetHDC());
                    }
                }
            }
        }

        public void KeyPress(object sender, KeyPressEventArgs e)
        {
            // empty
        }

        /**
         * @method KeyDown
         * 
         * @author Kevin Niehage
         * 
         * @param[in] event emitter
         * @param[in] event arguments
         * 
         * @brief Handles all key down events.
         */
        public void KeyDown(object sender, KeyEventArgs e)
        {
            if (FActive)
            {
                bool LChanged = false;

                if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.S))
                {
                    FController.SetKeyState(ControllerKey.KEY_DOWN, true);
                    LChanged = true;
                }
                if ((e.KeyCode == Keys.Left) || (e.KeyCode == Keys.A))
                {
                    FController.SetKeyState(ControllerKey.KEY_LEFT, true);
                    LChanged = true;
                }
                if ((e.KeyCode == Keys.Right) || (e.KeyCode == Keys.D))
                {
                    FController.SetKeyState(ControllerKey.KEY_RIGHT, true);
                    LChanged = true;
                }
                if ((e.KeyCode == Keys.Up) || (e.KeyCode == Keys.W))
                {
                    FController.SetKeyState(ControllerKey.KEY_UP, true);
                    LChanged = true;
                }
                if ((e.KeyCode == Keys.Return) || (e.KeyCode == Keys.Space))
                {
                    FController.SetKeyState(ControllerKey.KEY_JUMP, true);
                    LChanged = true;
                }

                if ((LChanged) && (!(FPlayReplay)) && (FClientSender != null))
                {
                    byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetController(FController, FTimeStamp)));
                    FClientSender.Send(LData, LData.Length, FServerEndPoint);
                }
            }
        }

        /**
         * @method KeyUp
         * 
         * @author Kevin Niehage
         * 
         * @param[in] event emitter
         * @param[in] event arguments
         * 
         * @brief Handles all key up events.
         */
        public void KeyUp(object sender, KeyEventArgs e)
        {
            if (FActive)
            {
				bool LChanged = false;

				if ((e.KeyCode == Keys.Down) || (e.KeyCode == Keys.S))
				{
					FController.SetKeyState(ControllerKey.KEY_DOWN, false);
					LChanged = true;
				}
				if ((e.KeyCode == Keys.Left) || (e.KeyCode == Keys.A))
				{
					FController.SetKeyState(ControllerKey.KEY_LEFT, false);
					LChanged = true;
				}
				if ((e.KeyCode == Keys.Right) || (e.KeyCode == Keys.D))
				{
					FController.SetKeyState(ControllerKey.KEY_RIGHT, false);
					LChanged = true;
				}
				if ((e.KeyCode == Keys.Up) || (e.KeyCode == Keys.W))
				{
					FController.SetKeyState(ControllerKey.KEY_UP, false);
					LChanged = true;
				}
                if ((e.KeyCode == Keys.Return) || (e.KeyCode == Keys.Space))
                {
                    FController.SetKeyState(ControllerKey.KEY_JUMP, false);
                    LChanged = true;
                }

                if ((LChanged) && (!(FPlayReplay)) && (FClientSender != null))
                {
					byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetController(FController, FTimeStamp)));
					FClientSender.Send(LData, LData.Length, FServerEndPoint);
				}
            }
        }

        public void MouseDown(object sender, MouseEventArgs e)
        {
            // empty
        }

        public void MouseUp(object sender, MouseEventArgs e)
        {
            // empty
        }

        /**
         * @method MouseMove
         * 
         * @author Kevin Niehage
         * 
         * @param[in] event emitter
         * @param[in] event arguments
         * 
         * @brief Handles all mouse move events.
         */
        public void MouseMove(object sender, MouseEventArgs e)
        {
            if (FActive)
            {
                GRControl LGRControl = (sender as GRControl);
                if (LGRControl != null)
                {
                    GR LGR = LGRControl.GetGR();
                    if (LGR != null)
                    {
                        Point lCursorPos = Cursor.Position;
                        lCursorPos.X -= FOwner.Left + LGRControl.Left + (LGRControl.Width / 2);
                        lCursorPos.Y -= FOwner.Top + LGRControl.Top + (LGRControl.Height / 2);

                        if ((Math.Abs(lCursorPos.X) > 10) || (Math.Abs(lCursorPos.Y) > 0))
                        {
                            FController.Rotation.X += 0.05f * (float)(lCursorPos.Y);
                            FController.Rotation.Y += 0.05f * (float)(lCursorPos.X);
                            FController.Rotation.Y %= 360;
                            if (FController.Rotation.Y < 0)
                            {
                                FController.Rotation.Y += 360;
                            }

                            if (FController.Rotation.X < -60)
                            {
                                FController.Rotation.X = -60;
                            }
                            else
                            {
                                if (FController.Rotation.X > 60)
                                {
                                    FController.Rotation.X = 60;
                                }
                            }
                            FController.Rotation.Z = 0.0F;

                            if ((!(FPlayReplay)) && (FClientSender != null))
                            {
                                byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetController(FController, FTimeStamp)));
                                FClientSender.Send(LData, LData.Length, FServerEndPoint);
                            }

                            Cursor.Position = new System.Drawing.Point(FOwner.Left + LGRControl.Left + (LGRControl.Width / 2),
                                                                       FOwner.Top + LGRControl.Top + (LGRControl.Height / 2));
                        }
                    }
                }
			}
		}

        public void MouseWheel(object sender, MouseEventArgs e)
        {
            // empty
        }
    }
}
