/**
 * @file ServerHandler.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class ServerHandler.
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Windows.Forms;

using Objects3D;

namespace Cubes3D
{
    /** 
     * @class ClientHandler
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for all functionality concerning the server.
     */
    public class ServerHandler
    {
        private Controller[] FControllers = null;
        private Player[]     FPlayers     = null;
        private ServerCube[] FServerCubes = null;

        private Thread FServerReceiver = null;
        private UdpClient FServerSender = null;
        private IPEndPoint FClientEndPoint = null;

        private bool FMap = true;

        private bool FActive      = false;
        private bool FGameRunning = false;
        private bool FKill        = false;

        private int FClientPort = 9999;
        private int FServerPort = 8888;

        private int FTickCount;

        private System.Windows.Forms.Timer FServerTimer = null;

        private bool FLocal;

        public ServerHandler(String AMapFile, bool ALocal)
        {
            FLocal = ALocal;

            int LTimeStamp = 0;

            if (File.Exists(AMapFile))
            {
                Network.SetServerCubes(File.ReadAllBytes(AMapFile), ref FServerCubes, ref LTimeStamp);
            }

            if (FLocal)
            {
                FClientEndPoint = new IPEndPoint(IPAddress.Loopback, FClientPort);
            }
            else
            {
                FClientEndPoint = new IPEndPoint(IPAddress.Broadcast, FClientPort);
            }

            FServerSender = new UdpClient();
            FServerSender.EnableBroadcast = (!(FLocal));

            FTickCount = System.Environment.TickCount;

            FServerReceiver = new Thread(new ThreadStart(StartServerReceiver));
            FServerReceiver.IsBackground = true; // simplify thread killing
            FServerReceiver.Start();

            FServerTimer = new System.Windows.Forms.Timer();
            FServerTimer.Interval = 10;
            FServerTimer.Tick += ServerTimerTick;
            FServerTimer.Enabled = true;
        }

        /**
         * @property Active
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the server is working.
         */
        public bool Active
        {
            get
            {
                return FActive;
            }

            set
            {
                FActive = value;

                if (!(FServerReceiver.IsAlive))
                {
                    FServerReceiver.Start();
                }
            }
        }

        /**
         * @property GameRunning
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the game is running.
         */
        public bool GameRunning
        {
            get
            {
                return FGameRunning;
            }

            set
            {
                FGameRunning = value;
            }
        }

        /**
         * @property Kill
         * 
         * @author Kevin Niehage
         * 
         * @brief Property defining whether the thread is about to be killed
         */
        public bool Kill
        {
            get
            {
                return FKill;
            }

            set
            {
                FKill = value;
            }
        }

        /**
         * @property PlayerCount
         * 
         * @author Kevin Niehage
         * 
         * @brief Property holding the current number of players
         */
        public byte PlayerCount
        {
            get
            {
                if (FPlayers != null)
                {
                    return (byte)FPlayers.Length;
                }
                else
                {
                    return 0;
                }
            }
        }

        /**
         * @method StartServerReceiver
         * 
         * @author Kevin Niehage
         * 
         * @brief Main method of the server UDP listener thread.
         */
        private void StartServerReceiver()
        {
            IPEndPoint LServerEndPoint = null;
            if (FLocal)
            {
                LServerEndPoint = new IPEndPoint(IPAddress.Loopback, FServerPort);
            }
            else
            {
                LServerEndPoint = new IPEndPoint(IPAddress.Any, FServerPort);
            }

            UdpClient LServerReceiver = new UdpClient(FServerPort);
            LServerReceiver.EnableBroadcast = (!(FLocal));

            Controller LController = null;

            byte[] LReceivedData;
            int LTickCount = 0; ;

            FActive = true;
            while (!(FKill))
            {
                if (FActive)
                {
                    LReceivedData = LServerReceiver.Receive(ref LServerEndPoint);
                    LReceivedData = Convert.FromBase64String(new ASCIIEncoding().GetString(LReceivedData));

                    if (FGameRunning)
                    {
                        if (Network.SetController(LReceivedData, ref LController, ref LTickCount))
                        {
                            if (LController.ID < FControllers.Length)
                            {
                                FControllers[LController.ID] = LController;

                                if (LController.ID < FPlayers.Length)
                                {
                                    FPlayers[LController.ID].Rotation = LController.Rotation;
                                }
                            }
                        }
                    }
                    else
                    {
                        bool LFound = false;
                        int LRandom = 0;
                        byte LID = 0;
                        if (Network.SetEnter(LReceivedData, ref LRandom, ref LTickCount))
                        {
                            LFound = false;
                            if (FPlayers != null)
                            {
                                foreach (Player LPlayer in FPlayers)
                                {
                                    LFound = (LPlayer.Random == LRandom);
                                    if (LFound)
                                    {
                                        LID = LPlayer.ID;
                                        break;
                                    }
                                }
                            }

                            if (!(LFound))
                            {
                                if (FPlayers == null)
                                {
                                    FPlayers = new Player[1];

                                    LID = 0;

                                    FPlayers[LID] = new Player();
                                    FPlayers[LID].ID = LID;
                                    FPlayers[LID].Random = LRandom;

                                    FControllers = new Controller[1];

                                    FControllers[LID] = new Controller();
                                    FControllers[LID].ID = LID;
                                }
                                else
                                {
                                    Player[] LPlayers = new Player[FPlayers.Length + 1];
                                    for (int LIndex = 0; LIndex < FPlayers.Length; LIndex++)
                                    {
                                        LPlayers[LIndex] = FPlayers[LIndex];
                                    }

                                    LID = (byte)(LPlayers.Length - 1);

                                    LPlayers[LID] = new Player();
                                    LPlayers[LID].ID = (byte)(LID);
                                    LPlayers[LID].Random = LRandom;

                                    FPlayers = LPlayers;

                                    Controller[] LControllers = new Controller[FControllers.Length + 1];
                                    for (int LIndex = 0; LIndex < FControllers.Length; LIndex++)
                                    {
                                        LControllers[LIndex] = FControllers[LIndex];
                                    }

                                    LControllers[LID] = new Controller();
                                    LControllers[LID].ID = LID;

                                    FControllers = LControllers;
                                }

                                if (FServerCubes != null)
                                {
                                    if (FServerCubes.Length > 0)
                                    {
                                        FPlayers[LID].Collided = true;
                                        FPlayers[LID].LastCube = FServerCubes[0];

                                        FPlayers[LID].Position.X = -FServerCubes[0].Cube.Pos.X;
                                        FPlayers[LID].Position.Y = FServerCubes[0].Cube.Pos.Y - Player.Scale - (FServerCubes[0].Cube.Scale.Y * FServerCubes[0].Cube.ScaleMultiplier);
                                        FPlayers[LID].Position.Z = -FServerCubes[0].Cube.Pos.Z;
                                    }
                                }
                            }

                            byte[] LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetAcknowledge(LRandom, LID, 0)));
                            FServerSender.Send(LData, LData.Length, FClientEndPoint);
                        }
                    }
                }
            }
        }

        /**
         * @method ServerTimerTick
         * 
         * @author Kevin Niehage
         * 
         * @param[in] event emitter
         * @param[in] event arguments
         *
         * @brief Method handling all intervaled events.
         */
        private void ServerTimerTick(object sender, EventArgs e)
        {
            byte[] LData;

            if (FGameRunning)
            {
                if (FMap)
                {
                    foreach (ServerCube LServerCube in FServerCubes)
                    {
                        LServerCube.Cube.Rot.X += LServerCube.RotSpeed.X;
                        LServerCube.Cube.Rot.X %= 360.0F;

                        LServerCube.Cube.Rot.Y += LServerCube.RotSpeed.Y;
                        LServerCube.Cube.Rot.Y %= 360.0F;

                        LServerCube.Cube.Rot.Z += LServerCube.RotSpeed.Z;
                        LServerCube.Cube.Rot.Z %= 360.0F;
                    }

                    LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetServerCubesAsCubes(FServerCubes, System.Environment.TickCount - FTickCount)));
                    FServerSender.Send(LData, LData.Length, FClientEndPoint);
                }
                else
                {
                    float LRadY = 0.0F;

                    bool LProceed = false;
                    foreach (Controller LController in FControllers)
                    {
                        if (LController.ID < FPlayers.Length)
                        {
                            if (!(FPlayers[LController.ID].Finished))
                            {
                                LProceed = true;

                                LRadY = Coordinate.DegToRad(LController.Rotation.Y);

                                if (LController.GetKeyState(ControllerKey.KEY_DOWN))
                                {
                                    FPlayers[LController.ID].Position.X += 5.0F * (float)Math.Sin(LRadY);
                                    FPlayers[LController.ID].Position.Z -= 5.0F * (float)Math.Cos(LRadY);
                                }
                                if (LController.GetKeyState(ControllerKey.KEY_LEFT))
                                {
                                    FPlayers[LController.ID].Position.X += 2.5F * (float)Math.Cos(LRadY);
                                    FPlayers[LController.ID].Position.Z += 2.5F * (float)Math.Sin(LRadY);
                                }
                                if (LController.GetKeyState(ControllerKey.KEY_RIGHT))
                                {
                                    FPlayers[LController.ID].Position.X -= 2.5F * (float)Math.Cos(LRadY);
                                    FPlayers[LController.ID].Position.Z -= 2.5F * (float)Math.Sin(LRadY);
                                }
                                if (LController.GetKeyState(ControllerKey.KEY_UP))
                                {
                                    FPlayers[LController.ID].Position.X -= 5.0F * (float)Math.Sin(LRadY);
                                    FPlayers[LController.ID].Position.Z += 5.0F * (float)Math.Cos(LRadY);
                                }

                                bool LCollision = false;
                                float LMovement = FPlayers[LController.ID].Gravity;
                                ServerCube LServerCube = null;
                                while (Math.Abs(LMovement) >= 1.0F)
                                {
                                    if (!(FPlayers[LController.ID].StartJump))
                                    {
                                        foreach (ServerCube LCube in FServerCubes)
                                        {
                                            float LNewY = 0.0F;
                                            LCollision = LCube.Cube.IsInBoundingBox(-FPlayers[LController.ID].Position.X, FPlayers[LController.ID].Position.Y, -FPlayers[LController.ID].Position.Z, Player.Scale, ref LNewY);
                                            if (LCollision)
                                            {
                                                FPlayers[LController.ID].Collided = true;
                                                FPlayers[LController.ID].LastCube = LCube;

                                                //FPlayers[LController.ID].Position.Y = LNewY;

                                                FPlayers[LController.ID].Gravity = 0.0F;
                                                FPlayers[LController.ID].Jump = false;

                                                LServerCube = LCube;
                                                break;
                                            }
                                        }
                                    }

                                    if (LCollision)
                                    {
                                        break;
                                    }
                                    else
                                    {
                                        if (LMovement > 0.0F)
                                        {
                                            FPlayers[LController.ID].Position.Y += 1.0F;
                                            LMovement -= 1.0F;
                                        }
                                        else
                                        {
                                            FPlayers[LController.ID].Position.Y -= 1.0F;
                                            LMovement += 1.0F;
                                        }
                                    }
                                }

                                if (FPlayers[LController.ID].StartJump)
                                {
                                    FPlayers[LController.ID].StartJump = false;
                                }

                                if (LController.GetKeyState(ControllerKey.KEY_JUMP))
                                {
                                    if (LCollision)
                                    {
                                        if (!(FPlayers[LController.ID].Jump))
                                        {
                                            FPlayers[LController.ID].Gravity = -10F;
                                            FPlayers[LController.ID].Jump = true;
                                            FPlayers[LController.ID].StartJump = true;
                                        }
                                    }
                                }

                                if (LCollision)
                                {
                                    if (FPlayers[LController.ID].LastCube == FServerCubes[FServerCubes.Length - 1])
                                    {
                                        FPlayers[LController.ID].Finished = true;
                                    }
                                }
                                else
                                {
                                    FPlayers[LController.ID].Gravity += Player.Acceleration / 25F;

                                    if (FPlayers[LController.ID].Position.Y > 1200F)
                                    {
                                        if (FPlayers[LController.ID].LastCube != null)
                                        {
                                            FPlayers[LController.ID].Gravity = 0F;
                                            FPlayers[LController.ID].Jump = false;
                                            FPlayers[LController.ID].StartJump = false;

                                            FPlayers[LController.ID].Position.X = -FPlayers[LController.ID].LastCube.Cube.Pos.X;
                                            FPlayers[LController.ID].Position.Y = FPlayers[LController.ID].LastCube.Cube.Pos.Y - Player.Scale - FPlayers[LController.ID].LastCube.Cube.Scale.Y * FPlayers[LController.ID].LastCube.Cube.ScaleMultiplier;
                                            FPlayers[LController.ID].Position.Z = -FPlayers[LController.ID].LastCube.Cube.Pos.Z;
                                        }
                                    }
                                }

                                if (FPlayers[LController.ID].Finished)
                                {
                                    LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetFinish(LController.ID, System.Environment.TickCount - FTickCount)));
                                    FServerSender.Send(LData, LData.Length, FClientEndPoint);
                                }
                            }
                        }
                    }

                    if (LProceed)
                    {
                        LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetPlayers(FPlayers, System.Environment.TickCount - FTickCount)));
                        FServerSender.Send(LData, LData.Length, FClientEndPoint);
                    }
                    else
                    {
                        LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetEnd(System.Environment.TickCount - FTickCount)));
                        FServerSender.Send(LData, LData.Length, FClientEndPoint);

                        FActive = false;
                        FGameRunning = false;
                        FKill = true;
                    }
                }

                FMap = (!(FMap));
            }
            else
            {
                LData = new ASCIIEncoding().GetBytes(Convert.ToBase64String(Network.GetNew(System.Environment.TickCount - FTickCount)));
                FServerSender.Send(LData, LData.Length, FClientEndPoint);
            }
        }
    }
}
