/**
 * @file Controller.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class Controller.
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Objects3D
{
    /** 
     * @class ControllerKey
     * 
     * @author Kevin Niehage
     * 
     * @brief Enumerator holding all possible game keys.
     */
    public enum ControllerKey
    {
        KEY_ATTACK,
        KEY_DOWN,
        KEY_JUMP,
        KEY_LEFT,
        KEY_RESET,
        KEY_RIGHT,
        KEY_UP
    }

    /** 
     * @class Controller
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for all the user input that is sent to the server.
     */
    public class Controller
    {
        protected byte FID = 0;

		protected Coordinate FRotation;
		protected byte[]     FStates;

        public Controller()
        {
            FID = 0;

			FRotation = new Coordinate();

            // has to match the number of different keys
            FStates = new byte[7]{0, 0, 0, 0, 0, 0, 0};
        }

        /**
         * @property ByteSize
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the length of the binary representation of Controller.
         */
        public static int ByteSize
        {
            get
            {
                return ((Coordinate.ByteSize * 1) + (sizeof(byte) * 8));
            }
        }

        /**
         * @method GetKeyState
         * 
         * @author Kevin Niehage
         * 
         * @param[in] key representer
         * 
         * @result indicator of the state of the passed key
         * 
         * @brief Method returning the key state of a key
         */
        public bool GetKeyState(ControllerKey AControllerKey)
        {
            return (FStates[(byte)AControllerKey] != 0);
        }

        /**
         * @method SetKeyState
         * 
         * @author Kevin Niehage
         * 
         * @param[in] key representer
         * @param[in] key state
         * 
         * @brief Method setting the key state of a key
         */
        public void SetKeyState(ControllerKey AControllerKey, bool AValue)
        {
            if ((AControllerKey >= ControllerKey.KEY_ATTACK) && (AControllerKey <= ControllerKey.KEY_UP))
            {
                if (AValue)
                {
                    FStates[(byte)AControllerKey] = 1;
                }
                else
                {
                    FStates[(byte)AControllerKey] = 0;
                }
            }
        }

        /**
         * @property ID
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the player ID.
         */
        public byte ID
        {
            get
            {
                return FID;
            }

            set
            {
                FID = value;
            }
        }

        /**
         * @property Rotation
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the current rotation of the player.
         */
        public Coordinate Rotation
		{
			get
			{
				return FRotation;
			}

			set
			{
                FRotation.Assign(value);
			}
		}

        /**
         * @method GetBytes
         * 
         * @author Kevin Niehage
         * 
         * @result binary representation of Controller instance
         * 
         * @brief Method returning binary representation of Controller instance.
         */
        public byte[] GetBytes()
        {
            byte[] LResult = new byte[ByteSize];

            byte[] LID = new byte[1]{FID};

			byte[] LRotation = FRotation.GetBytes();

            LID.CopyTo(LResult, (Coordinate.ByteSize * 0) + (sizeof(byte) * 0));
            LRotation.CopyTo(LResult, (Coordinate.ByteSize * 0) + (sizeof(byte) * 1));
            FStates.CopyTo(LResult, (Coordinate.ByteSize * 1) + (sizeof(byte) * 1));

            return LResult;
        }

        /**
         * @method SetBytes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ABytes array of byte containing the binary representation
         * @param[in] AOffset value defining from which position of the array the binary representation is to be read
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Method setting binary representation of Controller instance.
         */
        public bool SetBytes(byte[] ABytes, int AOffset)
        {
            bool LResult = false;

            if (((ABytes.Length - AOffset) >= ByteSize) && (AOffset >= 0))
            {
                FID = ABytes[AOffset++];

				FRotation.SetBytes(ABytes, AOffset);
				AOffset += Coordinate.ByteSize;

                for (int LIndex = AOffset; LIndex < ABytes.Length; LIndex++)
                {
                    FStates[LIndex - AOffset] = ABytes[LIndex];
                }

                LResult = true;
            }

            return LResult;
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] controller instance
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the values of the passed instance
         */
        public bool Assign(Controller AController)
        {
            bool LResult = false;

            if (AController != null)
            {
                ID = AController.ID;

                SetKeyState(ControllerKey.KEY_ATTACK, AController.GetKeyState(ControllerKey.KEY_ATTACK));
                SetKeyState(ControllerKey.KEY_DOWN, AController.GetKeyState(ControllerKey.KEY_DOWN));
                SetKeyState(ControllerKey.KEY_JUMP, AController.GetKeyState(ControllerKey.KEY_JUMP));
                SetKeyState(ControllerKey.KEY_LEFT, AController.GetKeyState(ControllerKey.KEY_LEFT));
                SetKeyState(ControllerKey.KEY_RESET, AController.GetKeyState(ControllerKey.KEY_RESET));
                SetKeyState(ControllerKey.KEY_RIGHT, AController.GetKeyState(ControllerKey.KEY_RIGHT));
                SetKeyState(ControllerKey.KEY_UP, AController.GetKeyState(ControllerKey.KEY_UP));

                LResult = FRotation.Assign(AController.Rotation);
            }

            return LResult;
        }
    }
}
