/**
 * @file Coordinate.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class Coordinate.
 */

using System;

namespace Objects3D
{
    /** 
     * @class Coordinate
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for a single 3-dimensional point (X, Y, Z).
     */
    public class Coordinate
    {
        protected event OnChangedHandler FOnAssigned;

        protected float FX;
        protected float FY;
        protected float FZ;

        public delegate void OnChangedHandler(Coordinate ACoordinate, float AOldX, float AOldY, float AOldZ);

        /**
         * @method RaiseOnAssigned
         * 
         * @author Kevin Niehage
         * 
         * @param[in] AOldX contains the old value of the property X
         * @param[in] AOldY contains the old value of the property Y
         * @param[in] AOldZ contains the old value of the property Z
         * 
         * @brief Wrapper for the call of event FOnAssigned().
         */
        protected virtual void RaiseOnAssigned(float AOldX, float AOldY, float AOldZ)
        {
            if (FOnAssigned != null)
            {
                FOnAssigned(this, AOldX, AOldY, AOldZ);
            }
        }

        public Coordinate()
        {
            FX = 0f;
            FY = 0f;
            FZ = 0f;
        }

        public Coordinate(float AX, float AY, float AZ)
        {
            FX = AX;
            FY = AY;
            FZ = AZ;
        }

        /**
         * @property ByteSize
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the length of the binary representation of Coordinate.
         */
        public static int ByteSize
        {
            get
            {
                return (sizeof(float) * 3);
            }
        }

        /**
         * @property OnAssigned
         * 
         * @author Kevin Niehage
         * 
         * @brief Wrapper around field FOnAssigned().
         */
        public OnChangedHandler OnAssigned
        {
            get
            {
                return FOnAssigned;
            }

            set
            {
                FOnAssigned = value;
            }
        }

        /**
         * @property X
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the X-value.
         */
        public float X
        {
            get
            {
                return FX;
            }

            set
            {
                Assign(value, FY, FZ);
            }
        }

        /**
         * @property Y
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the Y-value.
         */
        public float Y
        {
            get
            {
                return FY;
            }

            set
            {
                Assign(FX, value, FZ);
            }
        }

        /**
         * @property Z
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the Z-value.
         */
        public float Z
        {
            get
            {
                return FZ;
            }

            set
            {
                Assign(FX, FY, value);
            }
        }

        /**
         * @method GetBytes
         * 
         * @author Kevin Niehage
         * 
         * @result binary representation of Coordinate instance
         * 
         * @brief Method returning binary representation of Coordinate instance.
         */
        public byte[] GetBytes()
        {
            byte[] LResult = new byte[ByteSize];

            byte[] LX = BitConverter.GetBytes(FX);
            byte[] LY = BitConverter.GetBytes(FY);
            byte[] LZ = BitConverter.GetBytes(FZ);

            LX.CopyTo(LResult, (sizeof(float) * 0));
            LY.CopyTo(LResult, (sizeof(float) * 1));
            LZ.CopyTo(LResult, (sizeof(float) * 2));

            return LResult;
        }

        /**
         * @method SetBytes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ABytes array of byte containing the binary representation
         * @param[in] AOffset value defining from which position of the array the binary representation is to be read
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Method setting binary representation of Coordinate instance.
         */
        public bool SetBytes(byte[] ABytes, int AOffset)
        {
            bool LResult = false;

            if (((ABytes.Length - AOffset) >= ByteSize) && (AOffset >= 0))
            {
                Assign(BitConverter.ToSingle(ABytes, (sizeof(float) * 0) + AOffset),
                       BitConverter.ToSingle(ABytes, (sizeof(float) * 1) + AOffset),
                       BitConverter.ToSingle(ABytes, (sizeof(float) * 2) + AOffset));

                LResult = true;
            }

            return LResult;
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X coordinate value
         * @param[in] Y coordinate value
         * @param[in] Z coordinate value
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the passed values
         */
        public bool Assign(float AX, float AY, float AZ)
        {
            float LX = FX;
            float LY = FY;
            float LZ = FZ;

            FX = AX;
            FY = AY;
            FZ = AZ;

            RaiseOnAssigned(LX, LY, LZ);

            return true;
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] coordinate instance
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the values of the passed instance
         */
        public bool Assign(Coordinate ACoordinate)
        {
            bool LResult = false;

            if (ACoordinate != null)
            {
                LResult = Assign(ACoordinate.X, ACoordinate.Y, ACoordinate.Z);
            }

            return LResult;
        }

        /**
         * @method Equals
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X coordinate value
         * @param[in] Y coordinate value
         * @param[in] Z coordinate value
         * 
         * @result indicator if values are equal
         * 
         * @brief Checks whether the values of the current instance math then passed values
         */
        public bool Equals(float AX, float AY, float AZ)
        {
            return ((FX == AX) && (FY == AY) && (FZ == AZ));
        }

        /**
         * @method Equals
         * 
         * @author Kevin Niehage
         * 
         * @param[in] coordinate instance
         * 
         * @result indicator if values are equal
         * 
         * @brief Checks whether the values of the current instance math then values of the passed instance
         */
        public bool Equals(Coordinate ACoordinate)
        {
            bool LResult = false;

            if (ACoordinate != null)
            {
                LResult = Equals(ACoordinate.X, ACoordinate.Y, ACoordinate.Z);
            }

            return LResult;
        }

        /**
         * @method Rotate
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X rotate angle
         * @param[in] Y rotate angle
         * @param[in] Z rotate angle
         * @param[in] value for reverting the rotation
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Rotates the coordinate values around the center point
         */
        public bool Rotate(float AX, float AY, float AZ, bool APositive)
        {
            float LTempX = FX;
            float LTempY = FY;
            float LTempZ = FZ;

            if (APositive)
            {
                // rotation around Z axis
                FX = (float)(LTempX * Math.Cos(AZ) - LTempY * Math.Sin(AZ));
                FY = (float)(LTempX * Math.Sin(AZ) + LTempY * Math.Cos(AZ));
                FZ = LTempZ;

                // save temporarily
                LTempX = FX;
                LTempY = FY;
                LTempZ = FZ;

                // rotation around Y axis
                FX = (float)(LTempZ * Math.Sin(AY) + LTempX * Math.Cos(AY));
                FY = LTempY;
                FZ = (float)(LTempZ * Math.Cos(AY) - LTempX * Math.Sin(AY));

                // save temporarily
                LTempX = FX;
                LTempY = FY;
                LTempZ = FZ;

                // rotation around X axis
                FX = LTempX;
                FY = (float)(LTempY * Math.Cos(AX) - LTempZ * Math.Sin(AX));
                FZ = (float)(LTempY * Math.Sin(AX) + LTempZ * Math.Cos(AX));
            }
            else
            {
                // rotation around X axis
                FX = LTempX;
                FY = (float)(LTempY * Math.Cos(- AX) - LTempZ * Math.Sin(- AX));
                FZ = (float)(LTempY * Math.Sin(- AX) + LTempZ * Math.Cos(- AX));

                // save temporarily
                LTempX = FX;
                LTempY = FY;
                LTempZ = FZ;

                // rotation around Y axis
                FX = (float)(LTempZ * Math.Sin(- AY) + LTempX * Math.Cos(- AY));
                FY = LTempY;
                FZ = (float)(LTempZ * Math.Cos(- AY) - LTempX * Math.Sin(- AY));

                // save temporarily
                LTempX = FX;
                LTempY = FY;
                LTempZ = FZ;

                // rotation around Z axis
                FX = (float)(LTempX * Math.Cos(- AZ) - LTempY * Math.Sin(- AZ));
                FY = (float)(LTempX * Math.Sin(- AZ) + LTempY * Math.Cos(- AZ));
                FZ = LTempZ;
            }

            return true;
        }

        /**
         * @method Rotate
         * 
         * @author Kevin Niehage
         * 
         * @param[in] coordinate instance
         * @param[in] value for reverting the rotation
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Rotates the coordinate values around the center point
         */
        public bool Rotate(Coordinate ACoordinate, bool APositive)
        {
            bool LResult = false;

            if (ACoordinate != null)
            {
                LResult = Rotate(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, APositive);
            }

            return LResult;
        }

        /**
         * @method Scale
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X scaling value
         * @param[in] Y scaling value
         * @param[in] Z scaling value
         * @param[in] multiplier for scaling values
         * @param[in] value for reverting the scaling
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Scales the coordinate values 
         */
        public bool Scale(float AX, float AY, float AZ, float AMultiplier, bool APositive)
        {
            bool LResult = false;

            if (APositive)
            {
                FX *= AX * AMultiplier;
                FY *= AY * AMultiplier;
                FZ *= AZ * AMultiplier;

                LResult = true;
            }
            else
            {
                if ((AMultiplier != 0f) && (AX != 0f) && (AY != 0f) && (AZ != 0f))
                {
                    FX /= AX * AMultiplier;
                    FY /= AY * AMultiplier;
                    FZ /= AZ * AMultiplier;

                    LResult = true;
                }
            }

            return LResult;
        }

        /**
         * @method Scale
         * 
         * @author Kevin Niehage
         * 
         * @param[in] coordinate instance
         * @param[in] multiplier for scaling values
         * @param[in] value for reverting the scaling
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Scales the coordinate values 
         */
        public bool Scale(Coordinate ACoordinate, float AMultiplier, bool APositive)
        {
            bool LResult = false;

            if (ACoordinate != null)
            {
                LResult = Scale(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, AMultiplier, APositive);
            }

            return LResult;
        }

        /**
         * @method Translate
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X translation value
         * @param[in] Y translation value
         * @param[in] Z translation value
         * @param[in] value for reverting the translation
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Translates the coordinate
         */
        public bool Translate(float AX, float AY, float AZ, bool APositive)
        {
            if(APositive)
            {
                FX += AX;
                FY += AY;
                FZ += AZ;
            }
            else
            {
                FX -= AX;
                FY -= AY;
                FZ -= AZ;
            }

            return true;
        }

        /**
         * @method Translate
         * 
         * @author Kevin Niehage
         * 
         * @param[in] coordinate instance
         * @param[in] value for reverting the translation
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Translates the coordinate
         */
        public bool Translate(Coordinate ACoordinate, bool APositive)
        {
            bool LResult = false;

            if (ACoordinate != null)
            {
                LResult = Translate(ACoordinate.X, ACoordinate.Y, ACoordinate.Z, APositive);
            }

            return LResult;
        }

        /**
         * @method DegToRad
         * 
         * @author Kevin Niehage
         * 
         * @param[in] AAngle angle in degree
         * 
         * @result angle in radian measure
         * 
         * @brief Method returning radian measure value of passed degree angle.
         */
        public static float DegToRad(float AAngle)
        {
            return (float)((AAngle / 180f) * Math.PI);
        }

        /**
         * @method RadToDeg
         * 
         * @author Kevin Niehage
         * 
         * @param[in] AAngle angle in radian measure
         * 
         * @result angle in degree
         * 
         * @brief Method returning degree value of passed radian measure angle.
         */
        public static float RadToDeg(float AAngle)
        {
            return (float)((AAngle / Math.PI) * 180f);
        }
    }
}