/**
 * @file Cube.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class Cube.
 */

using System;

using GRV11;

namespace Objects3D
{
    /** 
     * @class Cube
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for all cube data needed to draw it.
     */
    public class Cube
    {
        protected byte FID = 0;

        protected Coordinate[,] FMatrix;
        protected Coordinate[] FPoints;

        protected Coordinate FPos = null;
        protected Coordinate FRot = null;
        protected Coordinate FScale = null;

        protected float FScaleMultiplier = 1.0F;

        protected void OnAssignedPos(Coordinate ACoordinate, float AOldX, float AOldY, float AOldZ)
        {
            if (!(FPos.Equals(AOldX, AOldY, AOldZ)))
            {
                foreach (Coordinate LIndex in FPoints)
                {
                    LIndex.Translate(AOldX, AOldY, AOldZ, false);

                    LIndex.Translate(FPos, true);
                }
            }
        }

        protected void OnAssignedRot(Coordinate ACoordinate, float AOldX, float AOldY, float AOldZ)
        {
            if (!(FRot.Equals(AOldX, AOldY, AOldZ)))
            {
                foreach (Coordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);
                    LIndex.Rotate(AOldX, AOldY, AOldZ, false);

                    LIndex.Rotate(FRot, true);
                    LIndex.Translate(FPos, true);
                }
            }
        }

        protected void OnAssignedScale(Coordinate ACoordinate, float AOldX, float AOldY, float AOldZ)
        {
            if (!(FScale.Equals(AOldX, AOldY, AOldZ)))
            {
                foreach (Coordinate LIndex in FPoints)
                {
                    LIndex.Translate(FPos, false);
                    LIndex.Rotate(FRot, false);
                    LIndex.Scale(AOldX, AOldY, AOldZ, FScaleMultiplier, false);

                    LIndex.Scale(FScale, FScaleMultiplier, true);
                    LIndex.Rotate(FRot, true);
                    LIndex.Translate(FPos, true);
                }
            }
        }

        public Cube()
        {
            FID = 0;

            FPoints = new Coordinate[8];
            FPoints[0] = new Coordinate(-0.5F, -0.5F, -0.5F);
            FPoints[1] = new Coordinate(+0.5F, -0.5F, -0.5F);
            FPoints[2] = new Coordinate(-0.5F, +0.5F, -0.5F);
            FPoints[3] = new Coordinate(-0.5F, -0.5F, +0.5F);
            FPoints[4] = new Coordinate(+0.5F, +0.5F, -0.5F);
            FPoints[5] = new Coordinate(+0.5F, -0.5F, +0.5F);
            FPoints[6] = new Coordinate(-0.5F, +0.5F, +0.5F);
            FPoints[7] = new Coordinate(+0.5F, +0.5F, +0.5F);

            FMatrix = new Coordinate[6, 4];

            // top
            FMatrix[0, 0] = FPoints[2];
            FMatrix[0, 1] = FPoints[6];
            FMatrix[0, 2] = FPoints[7];
            FMatrix[0, 3] = FPoints[4];

            // bottom
            FMatrix[1, 0] = FPoints[1];
            FMatrix[1, 1] = FPoints[5];
            FMatrix[1, 2] = FPoints[3];
            FMatrix[1, 3] = FPoints[0];

            // left
            FMatrix[2, 0] = FPoints[2];
            FMatrix[2, 1] = FPoints[0];
            FMatrix[2, 2] = FPoints[3];
            FMatrix[2, 3] = FPoints[6];

            // right
            FMatrix[3, 0] = FPoints[7];
            FMatrix[3, 1] = FPoints[5];
            FMatrix[3, 2] = FPoints[1];
            FMatrix[3, 3] = FPoints[4];

            // front
            FMatrix[4, 0] = FPoints[6];
            FMatrix[4, 1] = FPoints[3];
            FMatrix[4, 2] = FPoints[5];
            FMatrix[4, 3] = FPoints[7];

            // back
            FMatrix[5, 0] = FPoints[4];
            FMatrix[5, 1] = FPoints[1];
            FMatrix[5, 2] = FPoints[0];
            FMatrix[5, 3] = FPoints[2];

            FPos = new Coordinate(0.0F, 0.0F, 0.0F);
            FRot = new Coordinate(0.0F, 0.0F, 0.0F);
            FScale = new Coordinate(1.0F, 1.0F, 1.0F);

            FPos.OnAssigned += new Coordinate.OnChangedHandler(OnAssignedPos);
            FRot.OnAssigned += new Coordinate.OnChangedHandler(OnAssignedRot);
            FScale.OnAssigned += new Coordinate.OnChangedHandler(OnAssignedScale);

            FScaleMultiplier = 1.0F;
        }

        /**
         * @property ByteSize
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the length of the binary representation of Cube.
         */
        public static int ByteSize
        {
            get
            {
                return ((Coordinate.ByteSize * 3) + (sizeof(float) * 1) + (sizeof(byte) * 1));
            }
        }

        /**
         * @method GetBytes
         * 
         * @author Kevin Niehage
         * 
         * @result binary representation of Cube instance
         * 
         * @brief Method returning binary representation of Cube instance.
         */
        public byte[] GetBytes()
        {
            byte[] LResult = new byte[ByteSize];

            byte[] LID = new byte[1]{FID};

            byte[] LPos = FPos.GetBytes();
            byte[] LRot = FRot.GetBytes();
            byte[] LScale = FScale.GetBytes();

            byte[] LScaleMultiplier = BitConverter.GetBytes(FScaleMultiplier);

            LID.CopyTo(LResult, 0); 

            LPos.CopyTo(LResult, (Coordinate.ByteSize * 0) + (sizeof(byte) * LID.Length));
            LRot.CopyTo(LResult, (Coordinate.ByteSize * 1) + (sizeof(byte) * LID.Length));
            LScale.CopyTo(LResult, (Coordinate.ByteSize * 2) + (sizeof(byte) * LID.Length));

            LScaleMultiplier.CopyTo(LResult, (Coordinate.ByteSize * 3) + (sizeof(byte) * LID.Length));

            return LResult;
        }

        /**
         * @method SetBytes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ABytes array of byte containing the binary representation
         * @param[in] AOffset value defining from which position of the array the binary representation is to be read
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Method setting binary representation of Cube instance.
         */
        public bool SetBytes(byte[] ABytes, int AOffset)
        {
            bool LResult = false;

            if (((ABytes.Length - AOffset) >= ByteSize) && (AOffset >= 0))
            {
                FID = ABytes[AOffset];

                FPos.SetBytes(ABytes, (Coordinate.ByteSize * 0) + (sizeof(byte) * 1) + AOffset);
                FRot.SetBytes(ABytes, (Coordinate.ByteSize * 1) + (sizeof(byte) * 1) + AOffset);
                FScale.SetBytes(ABytes, (Coordinate.ByteSize * 2) + (sizeof(byte) * 1) + AOffset);

                ScaleMultiplier = BitConverter.ToSingle(ABytes, (Coordinate.ByteSize * 3) + (sizeof(byte) * 1) + AOffset);

                LResult = true;
            }

            return LResult;
        }

        /**
         * @property ID
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the cube ID.
         */
        public byte ID
        {
            get
            {
                return FID;
            }

            set
            {
                FID = value;
            }
        }

        /**
         * @property ScaleMultiplier
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the scaling multiplier.
         */
        public float ScaleMultiplier
        {
            get
            {
                return FScaleMultiplier;
            }

            set
            {
                if (value != FScaleMultiplier)
                {
                    foreach (Coordinate LIndex in FPoints)
                    {
                        LIndex.Translate(FPos, false);
                        LIndex.Rotate(FRot, false);
                        LIndex.Scale(FScale, FScaleMultiplier, false);

                        LIndex.Scale(FScale, value, true);
                        LIndex.Rotate(FRot, true);
                        LIndex.Translate(FPos, true);
                    }

                    FScaleMultiplier = value;
                }
            }
        }

        /**
         * @property Pos
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the position.
         */
        public Coordinate Pos
        {
            get
            {
                return FPos;
            }

            set
            {
                if (!(FPos.Equals(value)))
                {
                    foreach (Coordinate LIndex in FPoints)
                    {
                        LIndex.Translate(FPos, false);

                        LIndex.Translate(value, true);
                    }

                    FPos.Assign(value);
                }
            }
        }

        /**
         * @property Rot
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the rotation.
         */
        public Coordinate Rot
        {
            get
            {
                return FRot;
            }

            set
            {
                if (!(FRot.Equals(value)))
                {
                    foreach (Coordinate LIndex in FPoints)
                    {
                        LIndex.Translate(FPos, false);
                        LIndex.Rotate(FRot, false);

                        LIndex.Rotate(value, true);
                        LIndex.Translate(FPos, true);
                    }

                    FRot.Assign(value);
                }
            }
        }

        /**
         * @property Scale
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the scaling.
         */
        public Coordinate Scale
        {
            get
            {
                return FScale;
            }

            set
            {
                if (!(FScale.Equals(value)))
                {
                    foreach (Coordinate LIndex in FPoints)
                    {
                        LIndex.Translate(FPos, false);
                        LIndex.Rotate(FRot, false);
                        LIndex.Scale(FScale, FScaleMultiplier, false);

                        LIndex.Scale(value, FScaleMultiplier, true);
                        LIndex.Rotate(FRot, true);
                        LIndex.Translate(FPos, true);
                    }

                    FScale.Assign(value);
                }
            }
        }

        /**
         * @property Matrix
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the vertex matrix of the whole cube.
         */
        public Coordinate[,] Matrix
        {
            get
            {
                return FMatrix;
            }
        }

        /**
         * @property Points
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing all points that make up the cube.
         */
        public Coordinate[] Points
        {
            get
            {
                return FPoints;
            }
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] Cube instance
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the values of the passed instance
         */
        public bool Assign(Cube ACube)
        {
            bool LResult = false;

            if (ACube != null)
            {
                ID = ACube.ID;

                LResult = FPos.Assign(ACube.Pos);
                LResult = (LResult && FRot.Assign(ACube.Rot));
                LResult = (LResult && FScale.Assign(ACube.Scale));

                ScaleMultiplier = ACube.ScaleMultiplier;
            }

            return LResult;
        }

        /**
         * @method IsInBoundingBox
         * 
         * @author Kevin Niehage
         * 
         * @param[in] X position
         * @param[in] Y position
         * @param[in] Z position
         * @param[in] radius
         * @param[out] reference to new Y position
         * 
         * @result indicator if passed point is in bounding box
         * 
         * @brief Checks whether a point is in the boundin box of a cube
         */
        public bool IsInBoundingBox(float AX, float AY, float AZ, float ARadius, ref float ANewY)
        {
            bool LResult = false;

            float LXmax = FPoints[0].X;
            float LXmin = FPoints[0].X;
            float LYmax = FPoints[0].Y;
            float LYmin = FPoints[0].Y;
            float LZmax = FPoints[0].Z;
            float LZmin = FPoints[0].Z;

            for (int LIndex = 1; LIndex < FPoints.GetLength(0); LIndex++)
            {
                if (FPoints[LIndex].X < LXmin)
                {
                    LXmin = FPoints[LIndex].X;
                }
                if (FPoints[LIndex].X > LXmax)
                {
                    LXmax = FPoints[LIndex].X;
                }

                if (FPoints[LIndex].Y < LYmin)
                {
                    LYmin = FPoints[LIndex].Y;
                }
                if (FPoints[LIndex].Y > LYmax)
                {
                    LYmax = FPoints[LIndex].Y;
                }

                if (FPoints[LIndex].Z < LZmin)
                {
                    LZmin = FPoints[LIndex].Z;
                }
                if (FPoints[LIndex].Z > LZmax)
                {
                    LZmax = FPoints[LIndex].Z;
                }
            }

            LResult = (LXmin <= AX) && (LXmax >= AX);
            if (LResult)
            {
                LResult = LResult = (LZmin <= AZ) && (LZmax >= AZ);
                if (LResult)
                {
                    LResult = (LYmin - 0.5F <= AY + ARadius) && (LYmin + 0.5F >= AY + ARadius);
                    if (LResult)
                    {
                        ANewY = AY;
                    }
                }
            }


            return LResult;
        }

        /**
         * @method Draw
         * 
         * @author Kevin Niehage
         * 
         * @param[in] GR instance
         * 
         * @brief draws the cube via OpenGL
         */
        public void Draw(GR AGR)
        {
            if (AGR != null)
            {
                AGR.glBegin(GR.GL_QUADS);

                for (int LIndexA = 0; LIndexA < FMatrix.GetLength(0); LIndexA++)
                {
                    for (int LIndexB = 0; LIndexB < FMatrix.GetLength(1); LIndexB++)
                    {
                        AGR.glVertex3f(FMatrix[LIndexA, LIndexB].X, FMatrix[LIndexA, LIndexB].Y, FMatrix[LIndexA, LIndexB].Z);
                    }
                }

                AGR.glEnd();
            }
        }
    }
}
