/**
 * @file Network.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing Network class.
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Objects3D
{
    /** 
     * @class NEtwork
     * 
     * @author Kevin Niehage
     * 
     * @brief Method library for assembling and disassembling network packets.
     */
    public class Network
    {
        public static String KeyHeader = "KEY"; // KEY
        public static String MapHeader = "MAP"; // MAP
        public static String PlrHeader = "PLR"; // PLayeR
        public static String WldHeader = "WLD"; // WorLD

        public static String AckHeader = "ACK"; // ACKnowledged
        public static String NewHeader = "NEW"; // NEW
        public static String NtrHeader = "NTR"; // eNTeR

        public static String EndHeader = "END"; // END
        public static String FinHeader = "FIN"; // FINish

        /**
         * @method GetEnd
         * 
         * @author Kevin Niehage
         * 
         * @param[in] time stamp
         * 
         * @result binary representation of END package
         * 
         * @brief Generates END package from given parameters
         */
        public static byte[] GetEnd(int ATimeStamp) // write END record
        {
            byte[] LResult = new byte[(sizeof(int) * 1) + (sizeof(byte) * 3)];

            byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

            // write header
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'E';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'N';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'D';

            // write timestamp
            LTimeStamp.CopyTo(LResult, (sizeof(int) * 0) + (sizeof(byte) * 3));

            return LResult;
        }

        /**
         * @method GetFinish
         * 
         * @author Kevin Niehage
         * 
         * @param[in] player id of client
         * @param[in] time stamp
         * 
         * @result binary representation of FIN package
         * 
         * @brief Generates FIN package from given parameters
         */
        public static byte[] GetFinish(byte AID, int ATimeStamp) // write FIN record
        {
            byte[] LResult = new byte[(sizeof(int) * 1) + (sizeof(byte) * 4)];

            byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

            // write header
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'F';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'I';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'N';

            // write timestamp
            LTimeStamp.CopyTo(LResult, (sizeof(int) * 0) + (sizeof(byte) * 3));

            // write data
            LResult[(sizeof(int) * 1) + (sizeof(byte) * 3)] = AID;

            return LResult;
        }

        /**
         * @method SetEnd
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles END package
         */
        public static bool SetEnd(byte[] AData, ref int ATimeStamp) // read END record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(EndHeader))
                {
                    if (AData.Length >= sizeof(int) + 3)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetFinish
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to random identifier of client
         * @param[out] reference to player id of client
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles ACK package
         */
        public static bool SetFinish(byte[] AData, ref byte AID, ref int ATimeStamp) // read FIN record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(FinHeader))
                {
                    if (AData.Length >= (sizeof(int) * 1) + 4)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, (sizeof(int) * 0) + 3);

                        // read data
                        AID = AData[(sizeof(int) * 1) + 3];

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method GetAcknowledge
         * 
         * @author Kevin Niehage
         * 
         * @param[in] random identifier of client
         * @param[in] player id of client
         * @param[in] time stamp
         * 
         * @result binary representation of ACK package
         * 
         * @brief Generates ACK package from given parameters
         */
        public static byte[] GetAcknowledge(int ARandom, byte AID, int ATimeStamp) // write ACK record
        {
            byte[] LResult = new byte[(sizeof(int) * 2) + (sizeof(byte) * 4)];

            byte[] LRandom    = BitConverter.GetBytes(ARandom);
            byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

            // write header
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'A';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'C';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'K';

            // write timestamp
            LTimeStamp.CopyTo(LResult, (sizeof(int) * 0) + (sizeof(byte) * 3));

            // write data
            LRandom.CopyTo(LResult, (sizeof(int) * 1) + (sizeof(byte) * 3));
            LResult[(sizeof(int) * 2) + (sizeof(byte) * 3)] = AID;

            return LResult;
        }

        /**
         * @method GetNew
         * 
         * @author Kevin Niehage
         * 
         * @param[in] time stamp
         * 
         * @result binary representation of NEW package
         * 
         * @brief Generates NEW package from given parameters
         */
        public static byte[] GetNew(int ATimeStamp) // write NEW record
        {
            byte[] LResult = new byte[(sizeof(int) * 1) + (sizeof(byte) * 3)];

            byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

            // write header
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'N';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'E';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'W';

            // write timestamp
            LTimeStamp.CopyTo(LResult, (sizeof(int) * 0) + (sizeof(byte) * 3));

            return LResult;
        }

        /**
         * @method GetEnter
         * 
         * @author Kevin Niehage
         * 
         * @param[in] random identifier of client
         * @param[in] time stamp
         * 
         * @result binary representation of NTR package
         * 
         * @brief Generates NTR package from given parameters
         */
        public static byte[] GetEnter(int ARandom, int ATimeStamp) // write NTR record
        {
            byte[] LResult = new byte[(sizeof(int) * 2) + (sizeof(byte) * 3)];

            byte[] LRandom    = BitConverter.GetBytes(ARandom);
            byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

            // write header
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'N';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'T';
            LResult[(sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'R';

            // write timestamp
            LTimeStamp.CopyTo(LResult, (sizeof(int) * 0) + (sizeof(byte) * 3));

            // write data
            LRandom.CopyTo(LResult, (sizeof(int) * 1) + (sizeof(byte) * 3));

            return LResult;
        }

        /**
         * @method SetAcknowledge
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to random identifier of client
         * @param[out] reference to player id of client
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles ACK package
         */
        public static bool SetAcknowledge(byte[] AData, ref int ARandom, ref byte AID, ref int ATimeStamp) // read ACK record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(AckHeader))
                {
                    if (AData.Length >= (sizeof(int) * 2) + 4)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, (sizeof(int) * 0) + 3);

                        // read data
                        ARandom = BitConverter.ToInt32(AData, (sizeof(int) * 1) + 3);
                        AID     = AData[(sizeof(int) * 2) + 3];

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetNew
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles NEW package
         */
        public static bool SetNew(byte[] AData, ref int ATimeStamp) // read NEW record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(NewHeader))
                {
                    if (AData.Length >= sizeof(int) + 3)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetEnter
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to random identifier of client
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles NTR package
         */
        public static bool SetEnter(byte[] AData, ref int ARandom, ref int ATimeStamp) // read NTR record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(NtrHeader))
                {
                    if (AData.Length >= (sizeof(int) * 2) + 3)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, (sizeof(int) * 0) + 3);

                        // read data
                        ARandom = BitConverter.ToInt32(AData, (sizeof(int) * 1) + 3);

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method GetController
         * 
         * @author Kevin Niehage
         * 
         * @param[in] controller instance
         * @param[in] time stamp
         * 
         * @result binary representation of KEY package
         * 
         * @brief Generates KEY package from given parameters
         */
        public static byte[] GetController(Controller AController, int ATimeStamp) // write KEY records
        {
            byte[] LResult = null;

            if (AController != null)
            {
                LResult = new byte[(Controller.ByteSize * 1) + (sizeof(int) * 1) + (sizeof(byte) * 3)];

                byte[] LController = AController.GetBytes();
                byte[] LTimeStamp  = BitConverter.GetBytes(ATimeStamp);

                // write header
                LResult[(Controller.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'K';
                LResult[(Controller.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'E';
                LResult[(Controller.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'Y';

                // write timestamp
                LTimeStamp.CopyTo(LResult, (Controller.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 3));

                // write data
                LController.CopyTo(LResult, (Controller.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 3));
            }

            return LResult;
        }

        /**
         * @method GetCubes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] array of cube instances
         * @param[in] time stamp
         * 
         * @result binary representation of WLD package
         * 
         * @brief Generates WLD package from given parameters
         */
        public static byte[] GetCubes(Cube[] ACubes, int ATimeStamp) // write WorLD records
        {
            int    LPos    = 0;
            byte[] LResult = null;

            if (ACubes != null)
            {
                LResult = new byte[(Cube.ByteSize * ACubes.Length) + (sizeof(int) * 1) + (sizeof(byte) * 4)];

                byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

                // write header
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'W';
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'L';
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'D';

                // write timestamp
                LTimeStamp.CopyTo(LResult, (Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 3));

                // write data length
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 3)] = (byte)ACubes.Length;

                // write data
                LPos = (Cube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 4);
                foreach (Cube LCube in ACubes)
                {
                    LCube.GetBytes().CopyTo(LResult, LPos);
                    LPos += Cube.ByteSize;
                }
            }

            return LResult;
        }

        /**
         * @method GetPlayers
         * 
         * @author Kevin Niehage
         * 
         * @param[in] array of player instances
         * @param[in] time stamp
         * 
         * @result binary representation of PLR package
         * 
         * @brief Generates PLR package from given parameters
         */
        public static byte[] GetPlayers(Player[] APlayers, int ATimeStamp) // write PLayeR records
        {
            int LPos = 0;
            byte[] LResult = null;

            if (APlayers != null)
            {
                LResult = new byte[(Player.ByteSize * APlayers.Length) + (sizeof(int) * 1) + (sizeof(byte) * 4)];

                byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

                // write header
                LResult[(Player.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'P';
                LResult[(Player.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'L';
                LResult[(Player.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'R';

                // write timestamp
                LTimeStamp.CopyTo(LResult, (Player.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 3));

                // write data length
                LResult[(Player.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 3)] = (byte)APlayers.Length;

                // write data
                LPos = (Player.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 4);
                foreach (Player LPlayer in APlayers)
                {
                    LPlayer.GetBytes().CopyTo(LResult, LPos);
                    LPos += Player.ByteSize;
                }
            }

            return LResult;
        }

        /**
         * @method GetServerCubes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] array of server cube instances
         * @param[in] time stamp
         * 
         * @result binary representation of MAP package
         * 
         * @brief Generates MAP package from given parameters
         */
        public static byte[] GetServerCubes(ServerCube[] AServerCubes, int ATimeStamp) // write MAP records
        {
            int LPos = 0;
            byte[] LResult = null;

            if (AServerCubes != null)
            {
                LResult = new byte[(ServerCube.ByteSize * AServerCubes.Length) + (sizeof(int) * 1) + (sizeof(byte) * 4)];

                byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

                // write header
                LResult[(ServerCube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'M';
                LResult[(ServerCube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'A';
                LResult[(ServerCube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'P';

                // write timestamp
                LTimeStamp.CopyTo(LResult, (ServerCube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 3));

                // write data length
                LResult[(ServerCube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 3)] = (byte)AServerCubes.Length;

                // write data
                LPos = (ServerCube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 4);
                foreach (ServerCube LServerCube in AServerCubes)
                {
                    LServerCube.GetBytes().CopyTo(LResult, LPos);
                    LPos += ServerCube.ByteSize;
                }
            }

            return LResult;
        }

        /**
         * @method GetServerCubesAsCubes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] array of server cube instances
         * @param[in] time stamp
         * 
         * @result binary representation of WLD package
         * 
         * @brief Generates WLD package from given parameters
         */
        public static byte[] GetServerCubesAsCubes(ServerCube[] AServerCubes, int ATimeStamp) // write WorLD records
        {
            int LPos = 0;
            byte[] LResult = null;

            if (AServerCubes != null)
            {
                LResult = new byte[(Cube.ByteSize * AServerCubes.Length) + (sizeof(int) * 1) + (sizeof(byte) * 4)];

                byte[] LTimeStamp = BitConverter.GetBytes(ATimeStamp);

                // write header
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 0)] = (byte)'W';
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 1)] = (byte)'L';
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 2)] = (byte)'D';

                // write timestamp
                LTimeStamp.CopyTo(LResult, (Cube.ByteSize * 0) + (sizeof(int) * 0) + (sizeof(byte) * 3));

                // write data length
                LResult[(Cube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 3)] = (byte)AServerCubes.Length;

                // write data
                LPos = (Cube.ByteSize * 0) + (sizeof(int) * 1) + (sizeof(byte) * 4);
                foreach (ServerCube LServerCube in AServerCubes)
                {
                    LServerCube.Cube.GetBytes().CopyTo(LResult, LPos);
                    LPos += Cube.ByteSize;
                }
            }

            return LResult;
        }

        /**
         * @method SetController
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to controller
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles KEY package
         */
        public static bool SetController(byte[] AData, ref Controller AController, ref int ATimeStamp) // read KEY record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(KeyHeader))
                {
                    if (AData.Length >= Controller.ByteSize + sizeof(int) + 3)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        // read data
                        if (AController == null)
                        {
                            AController  = new Controller();
                        }
                        AController.SetBytes(AData, sizeof(int) + 3);

                        LResult = true;
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetCubes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to cube array
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles WLD package
         */
        public static bool SetCubes(byte[] AData, ref Cube[] ACubes, ref int ATimeStamp) // read WorLD records
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(WldHeader))
                {
                    if (AData.Length >= sizeof(int) + 4)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        // read data length
                        byte LLength = AData[sizeof(int) + 3];
                        if (AData.Length >= (Cube.ByteSize * LLength) + sizeof(int) + 4)
                        {
                            // read data
                            if ((ACubes == null) || (ACubes.Length != LLength))
                            {
                                ACubes = new Cube[LLength];

                                for (int LIndex = 0; LIndex < LLength; LIndex++)
                                {
                                    ACubes[LIndex] = new Cube();
                                }
                            }

                            int LPos = sizeof(int) + 4;
                            foreach (Cube LCube in ACubes)
                            {
                                LCube.SetBytes(AData, LPos);
                                LPos += Cube.ByteSize;
                            }

                            LResult = true;
                        }
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetPlayers
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to player array
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles PLR package
         */
        public static bool SetPlayers(byte[] AData, ref Player[] APlayers, ref int ATimeStamp) // read PLayeR record
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(PlrHeader))
                {
                    if (AData.Length >= sizeof(int) + 4)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        // read data length
                        byte LLength = AData[sizeof(int) + 3];
                        if (AData.Length >= (Player.ByteSize * LLength) + sizeof(int) + 4)
                        {
                            // read data
                            if ((APlayers == null) || (APlayers.Length != LLength))
                            {
                                APlayers = new Player[LLength];

                                for (int LIndex = 0; LIndex < LLength; LIndex++)
                                {
                                    APlayers[LIndex] = new Player();
                                }
                            }

                            int LPos = sizeof(int) + 4;
                            foreach (Player LPlayer in APlayers)
                            {
                                LPlayer.SetBytes(AData, LPos);
                                LPos += Player.ByteSize;
                            }

                            LResult = true;
                        }
                    }
                }
            }

            return LResult;
        }

        /**
         * @method SetServerCubes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] packet in binary representation
         * @param[out] reference to server cube array
         * @param[out] reference to time stamp
         * 
         * @result indicator if method worked correlty
         * 
         * @brief disassembles MAP package
         */
        public static bool SetServerCubes(byte[] AData, ref ServerCube[] AServerCubes, ref int ATimeStamp) // read MAP records
        {
            bool LResult = false;

            if (AData.Length >= 3)
            {
                // read header
                if (new ASCIIEncoding().GetString(AData, 0, 3).Equals(MapHeader))
                {
                    if (AData.Length >= sizeof(int) + 4)
                    {
                        // read timestamp
                        ATimeStamp = BitConverter.ToInt32(AData, 3);

                        byte LLength = AData[sizeof(int) + 3];
                        if (AData.Length >= (ServerCube.ByteSize * LLength) + sizeof(int) + 4)
                        {
                            // read data
                            if ((AServerCubes == null) || (AServerCubes.Length != LLength))
                            {
                                AServerCubes = new ServerCube[LLength];

                                for (int LIndex = 0; LIndex < LLength; LIndex++)
                                {
                                    AServerCubes[LIndex] = new ServerCube();
                                }
                            }

                            int LPos = sizeof(int) + 4;
                            foreach (ServerCube LServerCube in AServerCubes)
                            {
                                LServerCube.SetBytes(AData, LPos);
                                LPos += ServerCube.ByteSize;
                            }

                            LResult = true;
                        }
                    }
                }
            }

            return LResult;
        }
    }
}
