/**
 * @file Player.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class Player.
 */

using System;
using System.Collections.Generic;
using System.Text;

using GRV11;

namespace Objects3D
{
    /** 
     * @class Player
     * 
     * @author Kevin Niehage
     * 
     * @brief Container for all player data needed by the server.
     */
    public class Player
    {
        public static float Acceleration = 9.81F;
        public static byte  Scale = 15;

        protected static byte  FDetails = 24;

        protected bool  FCollided      = false;
        protected bool  FFinished = false;
        protected float FGravity       = 0;
        protected byte  FID            = 0;
        protected bool  FJump          = false;
        protected ServerCube FLastCube = null;
        protected int   FRandom        = 0;
        protected bool FStartJump = false;

        protected Coordinate FPosition;
        protected Coordinate FRotation;

        public Player()
        {
            FCollided = false;
            FFinished = false;
            FGravity  = 0;
            FID       = 0;
            FJump     = false;
            FLastCube = null;
            FRandom   = 0;
            FStartJump = false;

            FPosition = new Coordinate();
            FRotation = new Coordinate();
        }

        /**
         * @property ByteSize
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the length of the binary representation of Player.
         */
        public static int ByteSize
        {
            get
            {
                return ((Coordinate.ByteSize * 2) + (sizeof(byte) * 1));
            }
        }

        /**
         * @property Collided
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing whether a collision has happened before.
         */
        public bool Collided
        {
            get
            {
                return FCollided;
            }

            set
            {
                FCollided = value;
            }
        }

        /**
         * @property Finished
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing whehther the player has finished the map.
         */
        public bool Finished
        {
            get
            {
                return FFinished;
            }

            set
            {
                FFinished = value;
            }
        }

        /**
         * @property ID
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the player ID.
         */
        public byte ID
        {
            get
            {
                return FID;
            }

            set
            {
                FID = value;
            }
        }

        /**
         * @property Gravity
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the current gravity impact on the player.
         */
        public float Gravity
        {
            get
            {
                return FGravity;
            }

            set
            {
                FGravity = value;
            }
        }

        /**
         * @property Jump
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the current jump state of the player.
         */
        public bool Jump
        {
            get
            {
                return FJump;
            }

            set
            {
                FJump = value;
            }
        }

        /**
         * @property LastCube
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the last touched cube of the player.
         */
        public ServerCube LastCube
        {
            get
            {
                return FLastCube;
            }

            set
            {
                FLastCube = value;
            }
        }

        /**
         * @property Position
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the current position of the player.
         */
        public Coordinate Position
        {
            get
            {
                return FPosition;
            }

            set
            {
                FPosition.Assign(value);
            }
        }

        /**
         * @property Random
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the random identifier of the client.
         */
        public int Random
        {
            get
            {
                return FRandom;
            }

            set
            {
                FRandom = value;
            }
        }

        /**
         * @property Rotation
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the current rotation of the player.
         */
        public Coordinate Rotation
        {
            get
            {
                return FRotation;
            }

            set
            {
                FRotation.Assign(value);
            }
        }

        /**
         * @property StartJump
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing if a jump action just started.
         */
        public bool StartJump
        {
            get
            {
                return FStartJump;
            }

            set
            {
                FStartJump = value;
            }
        }

        /**
         * @method GetBytes
         * 
         * @author Kevin Niehage
         * 
         * @result binary representation of Player instance
         * 
         * @brief Method returning binary representation of Player instance.
         */
        public byte[] GetBytes()
        {
            byte[] LResult = new byte[ByteSize];

            byte[] LID = new byte[1]{FID};

            byte[] LPosition = FPosition.GetBytes();
            byte[] LRotation = FRotation.GetBytes();

            LID.CopyTo(LResult,(Coordinate.ByteSize * 0) + (sizeof(byte) * 0));
            LPosition.CopyTo(LResult,(Coordinate.ByteSize * 0) + (sizeof(byte) * 1));
            LRotation.CopyTo(LResult,(Coordinate.ByteSize * 1) + (sizeof(byte) * 1));

            return LResult;
        }

        /**
         * @method SetBytes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ABytes array of byte containing the binary representation
         * @param[in] AOffset value defining from which position of the array the binary representation is to be read
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Method setting binary representation of Player instance.
         */
        public bool SetBytes(byte[] ABytes, int AOffset)
        {
            bool LResult = false;

            if (((ABytes.Length - AOffset) >= ByteSize) && (AOffset >= 0))
            {
                FID = ABytes[(Coordinate.ByteSize * 0) + (sizeof(byte) * 0) + AOffset];
                FPosition.SetBytes(ABytes, (Coordinate.ByteSize * 0) + (sizeof(byte) * 1) + AOffset);
                FRotation.SetBytes(ABytes, (Coordinate.ByteSize * 1) + (sizeof(byte) * 1) + AOffset);

                LResult = true;
            }

            return LResult;
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] Player instance
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the values of the passed instance
         */
        public bool Assign(Player APlayer)
        {
            bool LResult = false;

            if (APlayer != null)            
            {
                Collided = APlayer.Collided;
                Finished = APlayer.Finished;
                Gravity = APlayer.Gravity;
                ID = APlayer.ID;
                Jump = APlayer.Jump;
                LastCube = APlayer.LastCube;
                Random = APlayer.Random;
                StartJump = APlayer.StartJump;

                LResult = FPosition.Assign(APlayer.Position);
                LResult = (LResult && FRotation.Assign(APlayer.Rotation));
            }

            return LResult;
        }

        /**
         * @method Draw
         * 
         * @author Kevin Niehage
         * 
         * @param[in] GR instance
         * @param[in] pointer to OpenGL Quadric instance
         * 
         * @brief draws the player via OpenGL
         */
        public void Draw(GR AGR, IntPtr AQuadric)
        {
            if ((AGR != null) && (AQuadric != null))
            {
                AGR.glTranslatef(-FPosition.X, -FPosition.Y, -FPosition.Z);

                //AGR.glEnable(GR.GL_LIGHTING);
                //AGR.glEnable(GR.GL_LIGHT0);

                AGR.gluSphere(AQuadric, Scale, FDetails, FDetails);
            }
        }
    }
}
