/**
 * @file ServerCube.cs
 * 
 * @author Kevin Niehage
 * 
 * @brief File containing the class ServerCube.
 */

using System;
using System.Collections.Generic;
using System.Text;

namespace Objects3D
{
    /** 
     * @class ServerCube
     * 
     * @author Kevin Niehage
     * 
     * @brief Cube class enhanced with data needed by the server.
     */
    public class ServerCube
    {
        protected Cube       FCube = null;
        protected Coordinate FRotSpeed = null;

        public ServerCube()
        {
            FCube     = new Cube();
            FRotSpeed = new Coordinate(0.0F, 0.0F, 0.0F);
        }

        /**
         * @property ByteSize
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the length of the binary representation of ServerCube.
         */
        public static int ByteSize
        {
            get
            {
                return ((Cube.ByteSize * 1) + (Coordinate.ByteSize * 1));
            }
        }

        /**
         * @method GetBytes
         * 
         * @author Kevin Niehage
         * 
         * @result binary representation of ServerCube instance
         * 
         * @brief Method returning binary representation of ServerCube instance.
         */
        public byte[] GetBytes()
        {
            byte[] LResult = new byte[ByteSize];

            byte[] LCube     = FCube.GetBytes();
            byte[] LRotSpeed = FRotSpeed.GetBytes();

            LCube.CopyTo(LResult, (Cube.ByteSize * 0));
            LRotSpeed.CopyTo(LResult, (Cube.ByteSize * 1));

            return LResult;
        }

        /**
         * @method SetBytes
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ABytes array of byte containing the binary representation
         * @param[in] AOffset value defining from which position of the array the binary representation is to be read
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Method setting binary representation of ServerCube instance.
         */
        public bool SetBytes(byte[] ABytes, int AOffset)
        {
            bool LResult = false;

            if (((ABytes.Length - AOffset) >= ByteSize) && (AOffset >= 0))
            {
                FCube.SetBytes(ABytes, (Cube.ByteSize * 0) + AOffset);
                FRotSpeed.SetBytes(ABytes, (Cube.ByteSize * 1) + AOffset);

                LResult = true;
            }

            return LResult;
        }

        /**
         * @property Cube
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the Cube instance.
         */
        public Cube Cube
        {
            get
            {
                return FCube;
            }

            set
            {
                FCube.Assign(value);
            }
        }

        /**
         * @property RotSpeed
         * 
         * @author Kevin Niehage
         * 
         * @brief Property containing the cube rotation speed.
         */
        public Coordinate RotSpeed
        {
            get
            {
                return FRotSpeed;
            }

            set
            {
                FRotSpeed.Assign(value);
            }
        }

        /**
         * @method Assign
         * 
         * @author Kevin Niehage
         * 
         * @param[in] ServerCube instance
         * 
         * @result indicator if method worked correctly
         * 
         * @brief Assigns the fields of the current instance the values of the passed instance
         */
        public bool Assign(ServerCube AServerCube)
        {
            bool LResult = false;

            if (AServerCube != null)
            {
                LResult = FCube.Assign(AServerCube.Cube);
                LResult = (LResult && FRotSpeed.Assign(AServerCube.RotSpeed));
            }

            return LResult;
        }
    }
}
